"""
Tic Tac Toe Player

This file assembles the functions our AI will need to make informed decisions about how to play tic tac toe.

Some of these functions are already implemented, but it's up to us to complete the rest of them!
"""


# The two lines below import "libraries" (code someone else has written) that we might use in our own code
import math
import copy

# The lines below define what X , O, and EMPTY should refer to. "None" is Python's keyword for "nothing", or "null".
X = "X"
O = "O"
EMPTY = None


##################################################################
# Implement the functions below to help our AI work as intended! #
##################################################################


def player(board):
    """
    Returns player who has the next turn on a board.
    """
    return None


def actions(board):
    """
    Returns set of all possible actions (i, j) available on the board.
    """
    return None


def result(board, action):
    """
    Returns the board that results from making move (i, j) on the board.
    """
    return None


def terminal(board):
    """
    Returns True if game is over, False otherwise.
    """
    return None


def utility(board):
    """
    Returns 1 if X has won the game, -1 if O has won, 0 otherwise.
    """
    return None


###################################################
# Functions below are already implemented for us! #
###################################################


def winner(board):
    """
    Returns the winner of the game, if there is one.
    """
    for i in range(len(board)):
        if check_vertical(board, i):
            return check_vertical(board, i)
        elif check_horizontal(board, i):
            return check_horizontal(board, i)

    if check_upward_diagonal(board):
        return check_upward_diagonal(board)

    if check_downward_diagonal(board):
        return check_downward_diagonal(board)


def minimax(board):
    """
    Returns the optimal action for the current player on the board.
    """
    if player(board) == "X":
        v = -100
        for potential_action in actions(board):
            new_v = minvalue(result(board, potential_action))
            if new_v > v:
                v = new_v
                action = potential_action
        return action
    elif player(board) == "O":
        v = 100
        for potential_action in actions(board):
            new_v = maxvalue(result(board, potential_action))
            if new_v < v:
                v = new_v
                action = potential_action
        return action


def check_horizontal(board, i):
    winners = {"X": 0, "O" : 0}

    for j in range(len(board)):
        if board[i][j] == "X":
            winners["X"] += 1
        elif board[i][j] == "O":
            winners["O"] += 1

    if winners["X"] == 3:
        return "X"
    elif winners["O"] == 3:
        return "O"
    else:
        return None


def check_vertical(board, j):
    winners = {"X": 0, "O" : 0}

    for i in range(len(board)):
        if board[i][j] == "X":
            winners["X"] += 1
        elif board[i][j] == "O":
            winners["O"] += 1

    if winners["X"] == 3:
        return "X"
    elif winners["O"] == 3:
        return "O"
    else:
        return None


def check_downward_diagonal(board):
    winners = {"X": 0, "O" : 0}

    for i in range(len(board)):
        if board[i][i] == "X":
            winners["X"] += 1
        elif board[i][i] == "O":
            winners["O"] += 1

    if winners["X"] == 3:
        return "X"
    elif winners["O"] == 3:
        return "O"
    else:
        return None


def check_upward_diagonal(board):
    winners = {"X": 0, "O" : 0}

    for i in range(len(board)):
        if board[len(board) - 1 - i][i] == "X":
            winners["X"] += 1
        elif board[len(board) - 1 - i][i] == "O":
            winners["O"] += 1

    if winners["X"] == 3:
        return "X"
    elif winners["O"] == 3:
        return "O"
    else:
        return None


def maxvalue(board):
    v = -100
    if terminal(board):
        return utility(board)

    for action in actions(board):
        v = max(v, minvalue(result(board, action)))
    return v


def minvalue(board):
    v = 100
    if terminal(board):
        return utility(board)

    for action in actions(board):
        v = min(v, maxvalue(result(board, action)))
    return v

def initial_state():
    """
    Returns starting state of the board.
    """
    return [[EMPTY, EMPTY, EMPTY],
            [EMPTY, EMPTY, EMPTY],
            [EMPTY, EMPTY, EMPTY]]